@@ -1,6 +1,6 @@
 // boehm.cc - interface between libjava and Boehm GC.
 
-/* Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004
+/* Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005
    Free Software Foundation
 
    This file is part of libgcj.
@@ -191,7 +191,11 @@ _Jv_MarkObj (void *addr, void *msp, void *msl, void *env)
 	      // static members.
 	      // Note that field->u.addr may be null if the class c is
 	      // JV_STATE_LOADED but not JV_STATE_PREPARED (initialized).
-	      if (JvFieldIsRef (field) && p && field->isResolved()) 
+	      // Note also that field->type could be NULL in some
+	      // situations, for instance if the class has state
+	      // JV_STATE_ERROR.
+	      if (field->type && JvFieldIsRef (field)
+		  && p && field->isResolved()) 
 		{
 		  jobject val = *(jobject*) p;
 		  p = (GC_PTR) val;
@@ -250,6 +254,9 @@ _Jv_MarkObj (void *addr, void *msp, void *msl, void *env)
 	  p = (GC_PTR) ic->interpreted_methods;
 	  MAYBE_MARK (p, mark_stack_ptr, mark_stack_limit, ic);
 
+	  p = (GC_PTR) ic->source_file_name;
+	  MAYBE_MARK (p, mark_stack_ptr, mark_stack_limit, ic);
+
 	  for (int i = 0; i < c->method_count; i++)
 	    {
 	      // The interpreter installs a heap-allocated trampoline
@@ -291,6 +298,8 @@ _Jv_MarkObj (void *addr, void *msp, void *msl, void *env)
 		    = (_Jv_InterpMethod *) ic->interpreted_methods[i];
 		  if (im)
 		    {
+                      p = (GC_PTR) im->line_table;
+                      MAYBE_MARK (p, mark_stack_ptr, mark_stack_limit, ic);
 		      p = (GC_PTR) im->prepared;
 		      MAYBE_MARK (p, mark_stack_ptr, mark_stack_limit, ic);
 		    }
